package core

import (
	"bytes"
	"fmt"
	"math/big"
	"sort"
	"time"
	"validator/tree"
	"validator/util"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	validatorv1 "github.com/odysseus/odysseus-protocol/gen/proto/go/validator/v1"
	log "github.com/sirupsen/logrus"
	"github.com/syndtr/goleveldb/leveldb/errors"
)

// CommitMST merkle sum tree 保存当天的workload
func (v *Validator) CommitMST(proofMap map[common.Address]*validatorv1.ValidatedProof) (root common.Hash, sum *big.Int, err error) {
	if len(proofMap) == 0 {
		return common.Hash{}, big.NewInt(0), nil
	}
	rawKeys := make([]common.Address, 0)
	keys := make([][]byte, len(proofMap))
	dbKey := make([]byte, 0)
	dbVal := make([]byte, 0)
	vals := make([]*big.Int, len(proofMap))
	for key := range proofMap {
		rawKeys = append(rawKeys, key)
	}
	sort.Sort(util.AddressSlice(rawKeys))
	for i, key := range rawKeys {
		dbKey = append(dbKey, rawKeys[i][:]...)
		keys[i] = rawKeys[i].Bytes()
		vals[i] = new(big.Int).SetUint64(proofMap[key].Workload)
		dbVal = append(append(dbVal, vals[i].Bytes()...), []byte(":")...)
	}

	st := time.Now()

	mstTree := tree.NewMerkleSumTree(keys, vals)

	rootNode := mstTree.GetRoot()

	err = v.lvdb.Put([]byte(fmt.Sprintf("mstroot:%s", v.date)), rootNode.Value.Hash.Bytes())
	if err != nil {
		log.Error(err)
		return
	}
	err = v.lvdb.Put([]byte(fmt.Sprintf("mstsum:%s", v.date)), rootNode.Value.BigValue.Bytes())
	if err != nil {
		log.Error(err)
		return
	}

	err = v.lvdb.Put([]byte(fmt.Sprintf("mstk:%s", v.date)), dbKey)
	if err != nil {
		return
	}
	err = v.lvdb.Put([]byte(fmt.Sprintf("mstv:%s", v.date)), dbVal)
	if err != nil {
		return
	}
	log.WithFields(log.Fields{
		"root": rootNode.Value.Hash.Hex(),
		"sum":  rootNode.Value.BigValue.String(),
		"cost": time.Since(st).String(),
	}).Info("commit MST root")

	v.Lock()
	v.mstTreeCache[v.date] = mstTree
	v.Unlock()

	return rootNode.Value.Hash, rootNode.Value.BigValue, nil
}

// CommitMT merkle tree 保存 获取到的总余额（已提取+未提取）
func (v *Validator) CommitMT(objects []*validatorv1.MinerObject) (root common.Hash, err error) {
	if len(objects) == 0 {
		return common.Hash{}, nil
	}
	merkleProofs := make(tree.Proofs, 0)
	dbProofs := make([]byte, 0)
	for _, object := range objects {
		bigBalance, _ := new(big.Int).SetString(object.Balance, 10)
		payload := append(common.HexToAddress(object.Miner).Bytes(), common.LeftPadBytes(bigBalance.Bytes(), 32)...)
		_proof := crypto.Keccak256Hash(payload)
		merkleProofs = append(merkleProofs, _proof)
		dbProofs = append(dbProofs, _proof[:]...)
	}

	mtTree, err := tree.NewMerkleTree(merkleProofs)
	if err != nil {
		return
	}
	root = mtTree.GetRoot()

	st := time.Now()
	err = v.lvdb.Put([]byte(fmt.Sprintf("mtroot:%s", v.date)), root.Bytes())
	if err != nil {
		return
	}
	log.WithFields(log.Fields{
		"k": fmt.Sprintf("mtroot:%s", v.date),
		"v": root.String(),
	}).Debug()

	err = v.lvdb.Put([]byte(fmt.Sprintf("mtk:%s", v.date)), dbProofs)
	if err != nil {
		return
	}
	log.WithFields(log.Fields{
		"k":        fmt.Sprintf("mtk:%s", v.date),
		"v_length": len(dbProofs),
	}).Debug()
	v.Lock()
	v.mtTreeCache[v.date] = mtTree
	v.Unlock()

	log.WithFields(log.Fields{
		"root": root.Hex(),
		"cost": time.Since(st).String(),
	}).Info("commit MT root")

	return
}

// LoadMerkleTree 从leveldb加载merkle tree
func (v *Validator) LoadMerkleTree(date string) (ok bool) {
	if date == "" {
		return false
	}
	if _, ok = v.mtTreeCache[date]; ok {
		return
	}

	merkleTreeKey := fmt.Sprintf("mtk:%s", date)
	data, err := v.lvdb.Get([]byte(merkleTreeKey))
	if err != nil {
		if err == errors.ErrNotFound {
			return
		}
		log.WithError(err).Error("failed to load merkle proof")
		return
	}

	log.WithFields(log.Fields{
		"key":    merkleTreeKey,
		"length": len(data),
	}).Debug("diskdb load merkle tree")

	proofs := make([]common.Hash, len(data)/32)
	for i := 0; i < len(data)/32; i++ {
		copy(proofs[i][:], data[i*32:(i+1)*32])
	}

	mTree, err := tree.NewMerkleTree(proofs)
	if err != nil {
		log.WithError(err).Error("failed to load merkle proof")
		return
	}
	v.Lock()
	v.mtTreeCache[date] = mTree
	v.Unlock()
	log.WithFields(log.Fields{
		"date": date,
		"root": mTree.GetRoot().Hex(),
	}).Debug("load merkle tree")
	return true
}

// LoadMerkleSumTree 从leveldb加载merkle sum tree
func (v *Validator) LoadMerkleSumTree(date string) (ok bool) {
	if date == "" {
		return false
	}
	if _, ok = v.mstTreeCache[date]; ok {
		return
	}

	merkleSumTreeKey := fmt.Sprintf("mstk:%s", date)
	keyData, err := v.lvdb.Get([]byte(merkleSumTreeKey))
	if err != nil {
		if err == errors.ErrNotFound {
			return
		}
		log.WithError(err).Error("failed to load merkle sum proof k")
		return
	}
	log.WithFields(log.Fields{
		"key":    merkleSumTreeKey,
		"length": len(keyData),
	}).Debug("diskdb load merkle sum proof key")

	// data = addr1:addr2
	rawKeys := make([][]byte, len(keyData)/20)
	for i := 0; i < len(keyData)/20; i++ {
		rawKeys[i] = make([]byte, 20)
		copy(rawKeys[i], keyData[i*20:(i+1)*20])
	}
	merkleSumTreeVal := fmt.Sprintf("mstv:%s", date)
	valData, err := v.lvdb.Get([]byte(merkleSumTreeVal))
	if err != nil {
		if err == errors.ErrNotFound {
			return
		}
		log.WithError(err).Error("failed to load merkle sum proof v")
		return
	}

	log.WithFields(log.Fields{
		"key":    merkleSumTreeVal,
		"length": len(valData),
	}).Debug("diskdb load merkle sum proof val")

	vals := bytes.Split(valData, []byte(":"))
	bigVals := make([]*big.Int, len(vals))
	for i := range vals {
		bigVals[i] = big.NewInt(0).SetBytes(vals[i])
	}

	mstTree := tree.NewMerkleSumTree(rawKeys, bigVals)
	v.Lock()
	v.mstTreeCache[date] = mstTree
	v.Unlock()
	log.WithFields(log.Fields{"date": date, "root": mstTree.GetRoot()}).Debug("load merkle sum tree")
	return true
}
